/***
 * Sangjog CLI
 * Copyright(c) 2022 Hirak Santra
 */
 'use strict';

const { spawn } = require('child_process');
const nodePath = require('path');
const Utils = require('../utils');
const fs = require('fs');
const CPManager = require('./cp-manager');

class AppRunner {

    static devOptions = {
        serverPID: null,
        complTime: null,
        firstLoaded: false,
        complCount: 0,
        cp: null,
        fileChanged: false,
        options: null
    }

    static isAppFileChanged = false;

    static getDevCMD(complCount) {
        return `node ./node_modules/ts-node/dist/bin.js -r ./node_modules/tsconfig-paths/register index.ts --type development --dev-count ${complCount}`;
    }

    static getBuildCMD(buildPath) {
        return `node ./node_modules/typescript/bin/tsc --outDir ${buildPath} -p .`
    }

    static getStartCMD() {
        return `node build/index.js --type production`;
    }

    static devWatchFileLists(fileChangeFn) {
        const filesPaths = ['index.ts', 'src/sangjog-app.json', 'src/Application.ts', 'src/app'];
        AppRunner.isAppFileChanged = false;
        for(const filesPath of filesPaths) {
            fs.watch(filesPath, { recursive: true }, (event, filename)=>{
                if(!AppRunner.isAppFileChanged) {
                    fileChangeFn({filesPath, event, filename});
                    AppRunner.isAppFileChanged = true;
                    setTimeout(()=>{
                        AppRunner.isAppFileChanged = false;
                    },20);
                }
            });
        }
    }

    static createDevCP(complCount) {
        const CMD = AppRunner.getDevCMD(complCount);
        const devCP = spawn(CMD, {
            stdio: ['inherit', 'pipe', 'inherit'],
            shell: true, 
            cwd: process.cwd()
        });
        AppRunner.devOptions.complTime = new Date().getTime();
        devCP.stdout.on('data', (data)=>{
            const txt = data.toString();
            if(txt.indexOf('SANGJOG_SERVER_PID::')!==-1 && !AppRunner.devOptions.serverPID) {
                const pid = txt.match(/(\d+)/);
                AppRunner.devOptions.serverPID = parseInt(pid);
            } else {
                process.stdout.write(data);
            }
        });
        devCP.on('error', (err) => {
            console.error(`SANGJOG_ERROR: ${err}`);
        });
        devCP.on('close', (code) => {
            if(code===0) {
                console.log("\x1b[36;1;1m%s\x1b[0m", `sangjog has successfully compiled the file..`);
            }
        });
        return devCP;
    }

    static runDevelopCP() {
        console.log(`please wait while compiling TS files....`)
        AppRunner.devOptions.cp = AppRunner.createDevCP(AppRunner.devOptions.complCount);
        AppRunner.devOptions.complCount++;
    }

    static tsFileChanged(options) {
        const { filesPath } = options;
        console.log(`change(EVENT_NAME:${options.event}) detected in **/${filesPath}`);
        AppRunner.devOptions.cp = AppRunner.createDevCP(AppRunner.devOptions.complCount);
        AppRunner.devOptions.complCount++;
        AppRunner.devOptions.cp.on("exit", AppRunner.runDevelopCP);
        CPManager.kill(AppRunner.devOptions.cp.pid);
        CPManager.kill(AppRunner.devOptions.serverPID);
        AppRunner.devOptions.serverPID = null;
    }

    static develop() {
        console.log(`Please wait while executing...`);
        AppRunner.runDevelopCP();
        this.devWatchFileLists((options)=>{
            if(!AppRunner.devOptions.firstLoaded) {
                return;
            }
            if(options.filesPath==='src/app') {
                options.filesPath = `${options.filesPath}/${options.filename.replace(/\\/g, '/')}`;
            }
            if(((new Date().getTime()) - AppRunner.devOptions.complTime) > 2000) {
                AppRunner.devOptions.fileChanged = true;
                AppRunner.devOptions.options = options;
                AppRunner.tsFileChanged(options);
            } else {
                AppRunner.devOptions.complTime = new Date().getTime();
                console.log(`too much repeatitive (EVENT_NAME:${options.event}) actions on the file. please try after few seconds....`);
            }
        });
        setTimeout(()=>{
            AppRunner.devOptions.firstLoaded = true;
        }, 200);
    }

    static start() {
        console.log(`Please wait while executing...`)
        const CMD = AppRunner.getStartCMD();
        CPManager.runCommand(CMD, {}, { close: (code)=>{
            if(code===0) {
                console.log("\x1b[32;1;1m%s\x1b[0m", `Welcome to Sangjog App!`);
            }
        }});
    }

    static build() {
        console.log(`Please wait while executing...`);
        const config = require(nodePath.join(process.cwd(), 'src/sangjog-app.json'));
        const buildPath = config.sangjog.buildPath;
        const CMD = AppRunner.getBuildCMD(buildPath);
        CPManager.runCommand(CMD, {}, { close: (code)=>{
            if(code===0) {
                console.log("\x1b[32;1;1m%s\x1b[0m", `SNG Build is success!`);
            }
        }});
    }

    static install() {
        console.log('Please wait while installing..');
        CPManager.runCommand('npm install', null, { close: (code)=>{
            if(code===0) {
                console.dir(`Installation is completed!`);
            }
        }});
    }

}

module.exports = AppRunner;
